"""ProtoTorch datasets"""

from .abstract import CSVDataset, NumpyDataset
from .sklearn import (
    Blobs,
    Circles,
    Iris,
    Moons,
    Random,
)
from .spiral import Spiral
from .tecator import Tecator
from .xor import XOR
